/*
Copyright (c) 1994, The Regents of the University of Colorado

Permission to use, copy, modify, distribute, and sell this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the University of
Colorado not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.  The University of Colorado makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

The University of Colorado disclaims all warranties with regard to this
software, including all implied warranties of merchantability and
fitness, in no event shall the University of Colorado be liable for any
special, indirect or consequential damages or any damages whatsoever
resulting from loss of use, data or profits, whether in an action of
contract, negligence or other tortious action, arising out of or in
connection with the use or performance of this software.
*/

#include <stdio.h>
#include <math.h>

/* ++pac
#include <vera2pov.h>
*/
#include "vera2pov.h"

#define Max(x,y) (((x) > (y)) ? (x) : (y))
#define Min(x,y) (((x) < (y)) ? (x) : (y))

/*
  ObjCtr() calculates the geometric center of a given object.
*/
int ObjCtr(ntriangles,triangles,ctr)
  unsigned long int	ntriangles;	/* number of triangles */
  float			*triangles;	/* ptr to flt arry of triangle vertices */
  float			ctr[3];		/* ptr to 1x3 flt arry */
{
unsigned long int	i;
short			j,k;

if (ntriangles == 0) return(FAILURE);
for (k = 0; k < 3; ctr[k++] = 0.0);
for (i = 0; i < ntriangles; i++)
  for (j = 0; j < 3; j++)
    for (k = 0; k < 3; k++) ctr[k] = ctr[k] + *(triangles++);
for (k = 0; k < 3; k++) ctr[k] = ctr[k]/(3*ntriangles);
return(SUCCESS);
}

/*
  ObjMaxMin() places the values for the maximum & minimum
  x-, y-, and z-coordinates in a given object.
*/
int ObjMaxMin(ntriangles,triangles,max,min)
  unsigned long int	ntriangles;	/* number of triangles */
  float			*triangles;	/* ptr to flt arry of triangle vertices */
  float			*max;		/* ptr to 1x3 flt arry */
  float			*min;		/* ptr to 1x3 flt arry */
{
unsigned long int	i;
short			j,k;

if (ntriangles == 0) return(FAILURE);
for (j = 0; j < 3; max[j] = min[j] = *(triangles+j++));
for (i = 0; i < ntriangles; i++) {
  for (j = 0; j < 3; j++)
    for (k = 0; k < 3; k++) {
      max[k] = Max(max[k],*triangles);
      min[k] = Min(min[k],*triangles);
      triangles++;
    }
}
return(SUCCESS);
}


/*
  WritePOV2File() writes to a PoV version 2.x-formatted
  file a union of triangles or smooth_triangles describing
  the object named obj_name.  Also written are the object's
  geometric center and x-, y-, and z-coordinate maxima and
  minima.
*/
int WritePOV2File(povfile,obj_name,
                  ntriangles,triangles,normals,mtx,
                  center,verbose,really_verbose)
  char			*povfile;	/* destination filename */
  char			*obj_name;	/* name of object */
  unsigned long int	ntriangles;	/* # of triangles */
  float			*triangles;	/* arry of triangle vertices */
  float			*normals;	/* arry of vertex normals */
  float			mtx[4][4];	/* 4x4 arry -- transformation mtx */
  short			center;		/* T/F: center the object about origin */
  short			verbose;	/* T/F: print runtime info */
  short			really_verbose;	/* T/F: print a lot of runtime info */
{
FILE			*fp;
unsigned long int	i;
short			j,k;
float			*t,*n,*f;
float			len,max[3],min[3],tpt[3],ctr[3],offset[3];
char			primitive[15];

for (j = 0; j < 3; offset[j++] = 0.0);

/* transform vertices and vertex normals */
if (mtx != NULL) {
  /* create copy of originial vertices */
  if ((t = (float *) calloc(3*3*ntriangles,sizeof(float))) == NULL)
    return(NOT_ENOUGH_MEMORY);
  for (i = 0; i < 3*3*ntriangles; *(t+i) = *(triangles+i++));
  if (normals) {
  /* create copy of originial normals */
    if ((n = (float *) calloc(3*3*ntriangles,sizeof(float))) == NULL)
      return(NOT_ENOUGH_MEMORY);
    for (i = 0; i < 3*3*ntriangles; *(n+i) = *(normals+i++));
  }
  else n = NULL;
  
  for (i = 0; i < ntriangles; i++) {
    for (j = 0; j < 3; j++) {
      Transform(t+(i*9)+(j*3),mtx,t+(i*9)+(j*3),FALSE);	/* rotation AND translation */
      if (normals) Transform(n+(i*9)+(j*3),mtx,n+(i*9)+(j*3),TRUE); /* rotate only -- no translation */
    }
  }
}
else {
  t = triangles;
  n = normals;
}

/* ensure normals are normalized */
if (n != NULL) {
  if (verbose) fprintf(stdout,"Renormalizing...%c",really_verbose? '\n':NULL);
  f = n;
  for (i = 0; i < ntriangles; i++) {
    if (really_verbose)
      fprintf(stdout," %d:  (%f, %f, %f)\t->\t",i+1,*f,*(f+1),*(f+2));
    len = 0.0;
    for (j = 0; j < 3; len += (*(f+j))*(*(f+j++)));
    len = sqrt(len);
    if (len == 0.0)
      return(ZERO_LEN_NORM);
    for (j = 0; j < 3; j++) *f = (*(f++))/len;
    if (really_verbose)
      fprintf(stdout,"(%f, %f, %f)\n",*(f-3),*(f-2),*(f-1));
  }
  strcpy(primitive,SMOOTH_TRIANGLE);
  if (verbose) fprintf(stdout,"done.\n");
}
else strcpy(primitive,TRIANGLE);

/* open file povfile */
if ((fp = fopen(povfile,"w")) == NULL) {
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(FILE_UNOPENED);
}

/* print header */
if (Header(fp,"// ") != SUCCESS) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}
if (fprintf(fp,"\n") < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}
if (fprintf(fp,"// Object Name:  %s\n\n",obj_name) < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}

/* print object center */
if (ObjCtr(ntriangles,t,ctr) == SUCCESS) {
  if (center) for (j = 0; j < 3; offset[j] = ctr[j++]);
  if (fprintf(fp,"#declare %s_Ctr = <%f,%f,%f>\n",
        obj_name,ctr[0]-offset[0],ctr[1]-offset[1],ctr[2]-offset[2]) < 0) {
    fclose(fp);
    if (t != triangles) free(t);
    if (n != NULL && n != normals) free(n);
    return(WRITE_ERROR);
  }
  for (j = 0; j < 3; j++) {
    if (fprintf(fp,"#declare %s_%cCtr = %f\n",obj_name,'X'+j,ctr[j]-offset[j]) < 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
    }
  }
}
else {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(FAILURE);
}
if (fprintf(fp,"\n") < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}

/* print object maxima & minima */
if (ObjMaxMin(ntriangles,t,max,min) == SUCCESS) {
  if (fprintf(fp,"#declare %s_Max = <%f,%f,%f>\n",
        obj_name,max[0]-offset[0],max[1]-offset[1],max[2]-offset[2]) < 0) {
    fclose(fp);
    if (t != triangles) free(t);
    if (n != NULL && n != normals) free(n);
    return(WRITE_ERROR);
  }
  for (j = 0; j < 3; j++) {
    if (fprintf(fp,"#declare %s_%cMax = %f\n",
          obj_name,'X'+j,max[j]-offset[j]) < 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
    }
  }
  if (fprintf(fp,"#declare %s_Min = <%f,%f,%f>\n",
        obj_name,min[0]-offset[0],min[1]-offset[1],min[2]-offset[2]) < 0) {
    fclose(fp);
    if (t != triangles) free(t);
    if (n != NULL && n != normals) free(n);
    return(WRITE_ERROR);
  }
  for (j = 0; j < 3; j++) {
    if (fprintf(fp,"#declare %s_%cMin = %f\n",obj_name,'X'+j,min[j]-offset[j]) < 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
    }
  }
}
else {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(FAILURE);
}
if (fprintf(fp,"\n") < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}

/* print union of triangles or smooth_triangles */
if (fprintf(fp,"#declare %s = union {\n",obj_name) < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}
for (i = 0; i < ntriangles; i++) {
  if (verbose) fprintf(stderr,"Writing triangle %d.\r",i+1);
  if (fprintf(fp," %s {",primitive) < 0) {
    fclose(fp);
    if (t != triangles) free(t);
    if (n != NULL && n != normals) free(n);
    return(WRITE_ERROR);
  }
  for (j = 0; j < 2; j++) {
    if (fprintf(fp,"<%f,%f,%f>,",
          (*t++)-offset[0],(*t++)-offset[1],(*t++)-offset[2]) < 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
    }
    if (n != NULL)
      if (fprintf(fp,"<%f,%f,%f>,",*n++,*n++,*n++) < 0) {
        fclose(fp);
        if (t != triangles) free(t);
        if (n != NULL && n != normals) free(n);
        return(WRITE_ERROR);
    }
  }
  if (fprintf(fp,"<%f,%f,%f>",
        (*t++)-offset[0],(*t++)-offset[1],(*t++)-offset[2]) < 0) {
    fclose(fp);
    if (t != triangles) free(t);
    if (n != NULL && n != normals) free(n);
    return(WRITE_ERROR);
  }
  if (normals != NULL) {
    if (fprintf(fp,",<%f,%f,%f>}\n",*n++,*n++,*n++)< 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
    }
  }
  else if (fprintf(fp,"}\n") < 0) {
      fclose(fp);
      if (t != triangles) free(t);
      if (n != NULL && n != normals) free(n);
      return(WRITE_ERROR);
  }
}
if (fprintf(fp,"}\n") < 0) {
  fclose(fp);
  if (t != triangles) free(t);
  if (n != NULL && n != normals) free(n);
  return(WRITE_ERROR);
}
if (verbose) fprintf(stdout,"\n");
fclose(fp);
if (t != triangles) free(t);
if (n != NULL && n != normals) free(n);
return(SUCCESS);
}

